import asyncio
import functools

import requests as fuck_aiohttp

from .base import SauceNaoHandler

class DanbooruHandler(SauceNaoHandler):
    engine_id = 9
    """"""

    tag_capable = True
    """"""

    source_capable = True
    """"""


    async def get_tag_data(self, data):
        loop = asyncio.get_event_loop()
        r = await loop.run_in_executor(None, functools.partial(fuck_aiohttp.get, "https://danbooru.donmai.us/posts/" + str(data["danbooru_id"]) + ".json"))
        j = r.json()
        return set(j["tag_string"].split())

    async def get_source_data(self, data):
        loop = asyncio.get_event_loop()
        r = await loop.run_in_executor(None, functools.partial(fuck_aiohttp.get, "https://danbooru.donmai.us/posts/" + str(data["danbooru_id"]) + ".json"))
        j = r.json()
        return set([j["source"]])


from retaggr.engines.base import Engine, ImageResult

# External modules
import asyncio
import requests as fuck_aiohttp
import functools

class Dummy(Engine):
    """A dummy engine that's only useful for testing.
    """
    host = "https://danbooru.donmai.us"
    download_required = False

    def __init__(self, fail):
        self.fail = fail

    async def search_image(self, url):
        if self.fail:
            raise Exception("Task failed (intentional)")
        else:
            return ImageResult(["test"], ["test"], "safe")


import asyncio
import functools

import requests as fuck_aiohttp

from .base import SauceNaoHandler

class GelbooruHandler(SauceNaoHandler):
    engine_id = 25
    """"""

    tag_capable = True
    """"""

    source_capable = True
    """"""


    async def get_tag_data(self, data):
        loop = asyncio.get_event_loop()
        r = await loop.run_in_executor(None, functools.partial(fuck_aiohttp.get, "https://gelbooru.com/index.php?page=dapi&s=post&q=index&json=1&id=" + str(data["gelbooru_id"])))
        j = r.json()[0]
        return set(j["tags"].split())

    async def get_source_data(self, data):
        loop = asyncio.get_event_loop()
        r = await loop.run_in_executor(None, functools.partial(fuck_aiohttp.get, "https://gelbooru.com/index.php?page=dapi&s=post&q=index&json=1&id=" + str(data["gelbooru_id"])))
        j = r.json()[0]
        return set([j["source"]])



from retaggr.engines.base import Engine, ImageResult
from retaggr.errors import NotAvailableSearchException

# External imports
import asyncio
import functools
import requests as fuck_aiohttp
from fake_useragent import UserAgent
from lxml import html

class Iqdb(Engine):
    """Reverse searches https://iqdb.org for images.

    This booru does not required images to be downloaded before searching.

    This method may have unexpected failures relating to HTML parsing. This is because IQDB does not officially offer an API and the raw HTML
    parsing is not functional if the page is not fully loaded.

    :param min_score: Minimum search match percentage needed.
    :type min_score: float
    """
    host = "https://iqdb.org"
    download_required = False

    def __init__(self, min_score):
        self.min_score = min_score
        self.ua = UserAgent()

    async def search_image(self, url):
        tags = []

        params = {"url" : url}
        loop = asyncio.get_event_loop()
        r = await loop.run_in_executor(None, functools.partial(fuck_aiohttp.post, self.host, headers={"User-Agent": self.ua.random}, params=params))

        doc = html.fromstring(r.text)
        tables = doc.xpath("//div[@id='pages']/div/table/tr/td")

        row = 6
        while row < len(tables):
            # Percent similair
            if (tables[row].text) is None: # pragma: no cover
                row = row + 6
                continue
            try:
                percent = float(str.split(tables[row].text)[0][:-1])
                # Create tags list
                tags_str = tables[row-2].xpath("//a/img")
                temp_tags = tags_str[0].get('alt').split("Tags: ", 1)[1]
                # IQDBs tag responses leave... something to be desired. To be clear,
                # Given the tag string "a,b, c, D"
                # These two lines should sanitize the tags to
                # ["a", "b", "c", "d"]
                tags = [x.lower().replace(",", " ").split() for x in temp_tags.split()]
                tags = [x for y in tags for x in y]

                if percent > self.min_score:
                    tags.extend(tags)
            except: # pragma: no cover
                pass
            row = row + 6

        return ImageResult(tags, [], None)

    async def search_tag(self, tag):
        """Reverse search the booru for tag data.
        """
        raise NotAvailableSearchException("This engine cannot search tags.")


import asyncio
import functools
import time

import requests as fuck_aiohttp

from .base import SauceNaoHandler
from retaggr.aiohttp_requests import requests

class E621Handler(SauceNaoHandler):
    engine_id = 29
    """"""

    tag_capable = True
    """"""

    source_capable = True
    """"""


    last_request = None

    def __init__(self, username, app_name, version):
        self.user_agent = {"User-Agent": f"{app_name}/{version} (by {username} on e621)"}

    async def rate_limit_wait(self):
        if self.last_request is not None: # pragma: no cover
            current_time = time.time()
            if current_time == self.last_request:
                asyncio.sleep(1)

    async def get_tag_data(self, data):
        await self.rate_limit_wait()
        r = await requests.get("https://e621.net/posts.json", headers=self.user_agent, params={"tags": "id:" + str(data["e621_id"])})
        j = await r.json()
        self.last_request = time.time()

        tags = set()
        for category in j["posts"][0]["tags"]:
            for tag in j["posts"][0]["tags"][category]:
                tags.add(tag)

        return tags

    async def get_source_data(self, data):
        await self.rate_limit_wait()
        r = await requests.get("https://e621.net/posts.json", headers=self.user_agent, params={"tags": "id:" + str(data["e621_id"])})
        j = await r.json()
        self.last_request = time.time()

        sources = set()
        for source in j["posts"][0]["sources"]:
            sources.add(source)

        return sources



from typing import Set

class SauceNaoHandler:
    """Base class to handle SauceNao engine results.
    """
    engine_id = None
    """The engine ID. Engine IDs can be located on https://saucenao.com/status.html."""

    tag_capable = False
    """This determines if the Handler has the ability to retrieve tags. 
    
    If this is false, :meth:`get_tag_data` may not necessarily exist."""

    source_capable = False
    """This determines if the Handler has the ability to retrieve additional source data. 
    
    If this is false, :meth:`get_source_data` may not necessarily exist."""

    async def get_tag_data(self, data) -> Set[str]: # pragma: no cover
        """Get all the tags matching the supplied data."""
        pass

    async def get_source_data(self, data) -> Set[str]: # pragma: no cover
        """Extract the source from the supplied data."""
        pass


import datetime
import asyncio
import functools

from retaggr.engines.base import Engine, ImageResult
from retaggr.engines.saucenao.handlers import DanbooruHandler, GelbooruHandler, E621Handler, KonachanHandler, YandereHandler
from retaggr.errors import NotAvailableSearchException, EngineCooldownException
import requests as fuck_aiohttp

class SauceNao(Engine):
    """Reverse searches the SauceNao API and then does additional matching.

    This booru does not require images to be downloaded before searching.

    This API is subject to rate limits.

    :param api_key: SauceNao API key. You can get this by registering an account on saucenao.com
    :type api_key: str
    :param test_mode: Enable test mode. Test mode is unique in that it does not need an API key, but it only works on one URL.
    """
    host = "https://saucenao.com"
    download_required = False

    def __init__(self, api_key, test_mode=False):
        self.api_key = api_key
        self.handlers = {
            DanbooruHandler.engine_id : DanbooruHandler(),
            GelbooruHandler.engine_id : GelbooruHandler(),
            KonachanHandler.engine_id : KonachanHandler(),
            YandereHandler.engine_id : YandereHandler(),
        }
        self.test_mode = test_mode

    def enable_e621(self, username, app_name, version):
        """Enable the E621 parser. This allows for looking up tag information on E621.
        
        :param username: An E621 username.
        :type username: str
        :param app_name: The name of the application.
        :type app_name: str
        :param version: The version of the appliation.
        :type version: str
        """
        self.handlers[E621Handler.engine_id] = E621Handler(username, app_name, version)

    async def search_image(self, url):
        request_url = "https://saucenao.com/search.php"
        params = {
            "db": "999", # No clever bitmasking -> need help with how to do that.
            "api_key": self.api_key,
            "output_type": "2", # 2 is the JSON API,
            "url": url
        }

        if self.test_mode:
            params = {
                "db": "999",
                "output_type": "2",
                "testmode": "1",
                "numres": "16",
                "url": "http://saucenao.com/images/static/banner.gif"
            }

        loop = asyncio.get_event_loop()
        r = await loop.run_in_executor(None, functools.partial(fuck_aiohttp.get, request_url, params=params))
        j = r.json()
        if r.status_code == 200:
            return await self.index_parser(j)
        elif r.status_code == 429: # pragma: no cover
            raise EngineCooldownException()

    async def search_tag(self, tag):
        raise NotAvailableSearchException("This engine cannot search tags.")

    async def index_parser(self, json):
        """Parse the output from a succesful saucenao search to retrieve data from specific indexes.
        
        :param json: JSON output from the API.
        :type json: dict
        :return: Dictionary containing data that matches the output for :meth:`SauceNao.search_image_source`
        :rtype: ImageResult
        """
        base_similarity = json["header"]["minimum_similarity"] # Grab the minimum similarity saucenao advises, going lower is generally gonna give false positives.

        # Below we cast the _entry_ similarity to a float since somehow it's stored as an str.
        # Damn API inaccuracy
        valid_results = [entry for entry in json["results"] if float(entry["header"]["similarity"]) > base_similarity]

        # Test mode similarity override
        if self.test_mode:
            valid_results = json["results"]

        # Kinda looks stupid, but whatever.
        loop = asyncio.get_event_loop()
        source = set()
        tags = set()
        for entry in valid_results:
            if "ext_urls" in entry["data"]: # Some of these responses dont have ext_url...
                for url in entry["data"]["ext_urls"]:
                    source.add(url)
            handler = self.handlers.get(entry["header"]["index_id"], None)
            if handler:
                if handler.tag_capable:
                    tags.update(await handler.get_tag_data(entry["data"]))
                if handler.source_capable:
                    source.update(await handler.get_source_data(entry["data"]))

        return ImageResult(tags, source, None)



import asyncio
import functools

import requests as fuck_aiohttp

from .base import SauceNaoHandler

class YandereHandler(SauceNaoHandler):
    engine_id = 12
    """"""

    tag_capable = True
    """"""

    source_capable = False
    """"""

    async def get_tag_data(self, data):
        loop = asyncio.get_event_loop()
        r = await loop.run_in_executor(None, functools.partial(fuck_aiohttp.get, "https://yande.re/post.json", params={"tags": "id:" + str(data["yandere_id"])}))
        j = r.json()
        return set(j[0]["tags"].split())


import datetime
import asyncio
from collections import namedtuple

ImageResult = namedtuple("ImageResult", ["tags", "source", "rating"])
"""The response from the engine.

.. py:attribute:: tags

    The tags the engine has located.

.. py:attribute:: source

    The sources that have been found for the image.

.. py:attribute:: rating

    The rating on the image.

"""


class Engine:
    """Base class for an engine.
    
    All the engine classes must derive from this class.

    :ivar host: The base URL for the reverse image domain. This is not an API endpoint, but can be a link to IQDB or something similar.
    :vartype host: str
    :ivar download_required: Determines if the ``search_image()`` function will download the image to be searched beforehand or not.
    :vartype download_required: bool
    :ivar datetime.datetime ~.last_request: Optional. Instance variable that can be set to make use of :meth:`Booru.sleep_until_ratelimit`. Update after making a request. If you just use the library, this should be done by the engine so there's no need to manually set it.
    """
    host = None
    download_required = False

    def __init__(self): # pragma: no cover
        self.last_request = None
        raise NotImplementedError("Expand this method to include all needed keys.")

    async def search_tag(self, tag): # pragma: no cover
        """Reverse search the booru for tag data.
        """
        raise NotImplementedError("Expand this method to include the logic needed to reverse search.")

    async def search_image(self, url): # pragma: no cover
        """Reverse search the engine for ``url``.

        :param str url: URL to search
        :rtype: ImageResult
        """
        raise NotImplementedError("Expand this method to include the logic needed to reverse search.")


from retaggr.engines.base import Engine, ImageResult

# External modules
import asyncio
import requests as fuck_aiohttp
import functools

class Danbooru(Engine):
    """Reverse searches https://danbooru.donmai.us for images.
    
    This booru does not required images to be downloaded before searching.

    :param username: The danbooru username you wish to use.
    :type username: str
    :param api_key: A danbooru API key. You can obtain one from your profile.
    :type api_key: str
    :param min_score: Minimum search match percentage needed.
    :type min_score: float
    """
    host = "https://danbooru.donmai.us"
    download_required = False

    def __init__(self, username, api_key, min_score):
        self.min_score = min_score
        self.username = username
        self.api_key = api_key

    async def search_image(self, url):
        tags = []
        source = []
        rating = None

        iqdb_url = self.host + "/iqdb_queries.json"
        loop = asyncio.get_event_loop()
        r = await loop.run_in_executor(None, functools.partial(fuck_aiohttp.get, iqdb_url, params={"url":url}, auth=(self.username, self.api_key)))
        json = r.json()
        if 'success' in json:
            if not json['success']: # pragma: no cover
                return ImageResult(tags, source, rating)

        if len(json) > 0:
            if json[0]['score'] > self.min_score:
                tags = json[0]["post"]["tag_string"].split()
                source.append(json[0]["post"]["source"])
                rating = json[0]["post"]["rating"]
        return ImageResult(tags, source, rating)


from .core import ReverseSearch, ReverseResult
from .config import ReverseSearchConfig
from .errors import MissingAPIKeysException, NotAValidEngineException, NotAvailableSearchException, EngineCooldownException, EngineIsDown
from .engines import ImageResult

from collections import namedtuple

VersionInfo = namedtuple('VersionInfo', 'major minor micro releaselevel serial')
version_info = VersionInfo(major=3, minor=1, micro=0, releaselevel="final", serial=0)

__version__ = "{}.{}.{}".format(version_info.major, version_info.minor, version_info.micro)


from retaggr.engines.base import Engine, ImageResult
from retaggr.errors import NotAvailableSearchException

# External imports
import hashlib
from retaggr.aiohttp_requests import requests
import xml.etree.ElementTree as ET

class Paheal(Engine):
    """Reverse searches https://rule34.paheal.net for images.

    This booru does require images to be downloaded before searching.
    """
    host = None
    download_required = True

    def __init__(self):
        pass

    async def search_image(self, url):
        tags = []
        source = []

        m = hashlib.md5()
        r = await requests.get(url)
        async for data in r.content.iter_chunked(8192):
            m.update(data)
        md5_hash = m.hexdigest()
        paheal_request = await requests.get(f"http://rule34.paheal.net/api/danbooru/find_posts?md5={md5_hash}")
        xml_tree = ET.fromstring(await paheal_request.text())

        for post in xml_tree:
            for tag in post.attrib["tags"].split(): 
                tags.append(tag.lower())
            source.append(post.attrib["source"])
        return ImageResult(tags, source, None)

    async def search_tag(self, tag):
        """Reverse search the booru for tag data.
        """
        raise NotAvailableSearchException("This engine cannot search tags.")


# MIT License
#
# Copyright (c) 2018 Max Zheng
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
import aiohttp
import functools

# Patch ClientResponse.read to release immediately after read so we don't need to worry about that / use context manager
_read_only = aiohttp.client_reqrep.ClientResponse.read
async def _read_and_release(self):  # noqa
    try:
        data = await _read_only(self)
    finally:
        self.close()

    return data
aiohttp.client_reqrep.ClientResponse.read = _read_and_release


class Requests:
    """ Thin wrapper for aiohttp.ClientSession with Requests simplicity """
    def __init__(self, *args, **kwargs):
        self._session_args = (args, kwargs)
        self._session = None

    @property
    def session(self):
        """ An instance of aiohttp.ClientSession """
        if not self._session or self._session.closed or self._session.loop.is_closed():
            self._session = aiohttp.ClientSession(*self._session_args[0], **self._session_args[1])
        return self._session

    def __getattr__(self, attr):
        if attr.upper() in aiohttp.hdrs.METH_ALL:
            @functools.wraps(self.session._request)
            def session_request(*args, **kwargs):
                """
                This ensures `self.session` is always called where it can check the session/loop state so can't use
                functools.partials as monkeypatch seems to do something weird where __getattr__ is only called once for
                each attribute after patch is undone
                """
                return self.session._request(attr.upper(), *args, **kwargs)

            return session_request
        else:
            return super().__getattribute__(attr)

    def close(self):
        """
        Close aiohttp.ClientSession.

        This is useful to be called manually in tests if each test when each test uses a new loop. After close, new
        requests will automatically create a new session.

        Note: We need a sync version for `__del__` and `aiohttp.ClientSession.close()` is async even though it doesn't
        have to be.
        """
        if self._session:
            if not self._session.closed:
                # Older aiohttp does not have _connector_owner
                if not hasattr(self._session, '_connector_owner') or self._session._connector_owner:
                    self._session._connector.close()
                self._session._connector = None
            self._session = None

    def __del__(self):
        self.close()


requests = Requests()



from setuptools import setup, find_packages
import subprocess

setup(
    name="retaggr",
    version="3.1.0",
    url="https://github.com/noirscape/retaggr",
    license="LGPLv3",
    description="Reverse image searching utility for images.",
    long_description=open("README.md").read(),
    long_description_content_type='text/markdown',
    author="Valentijn 'noirscape' V.",
    author_email="neko@catgirlsin.space",
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "License :: OSI Approved :: GNU Lesser General Public License v3 (LGPLv3)",
        "Programming Language :: Python :: 3",
    ],
    keywords="reverse image search booru",
    project_urls={
        "Source": "https://github.com/noirscape/retaggr",
        "Tracker": "https://github.com/noirscape/retaggr/issues",
        "Documentation": "https://retaggr.rtfd.org"
    },
    packages=find_packages('src'),
    package_dir={'':'src',},
    install_requires=[
        "lxml",
        "fake-useragent",
        "requests",
        "aiohttp"
    ]
)



# stdlib
import asyncio
from collections import namedtuple
import logging
import traceback

# Config
from retaggr.config import ReverseSearchConfig

# Engines
from retaggr.engines.base import ImageResult
from retaggr.engines.danbooru import Danbooru
from retaggr.engines.iqdb import Iqdb
from retaggr.engines.paheal import Paheal
from retaggr.engines.saucenao import SauceNao
from retaggr.engines.dummy import Dummy

# Exceptions
from retaggr.errors import MissingAPIKeysException, NotAValidEngineException, EngineCooldownException

ReverseResult = namedtuple("ReverseResult", ["tags", "source", "rating"])
"""The response from a reverse image search. All attributes are Sets.

.. py:attribute:: tags

    The tags the engine has located.

.. py:attribute:: source

    All the sources that have been found for the image.

.. py:attribute:: rating

    The rating on the image.

"""

# Set up logger
logger = logging.getLogger(__name__)

class ReverseSearch:
    r"""Core class used for Reverse Searching. 
    
    This class can only be instantiated with a :class:`ReverseSearchConfig` instance.

    All listed methods can only be ran from an asynchronous context.

    :ivar accessible_engines: The accessible boorus from the passed in configuration object.
    :param config: The config object.
    :type config: ReverseSearchConfig
    :param test_mode: Enable test mode. Test mode adds two dummy engines that can fail as well as return some very basic values.
    :type test_mode: bool
    """
    _all_engines = [
        "danbooru",
        "e621",
        "iqdb",
        "paheal",
        "saucenao"
        ]

    def __init__(self, config, test_mode=False):
        self.config = config
        self.accessible_engines = {}

        if hasattr(self.config, "min_score"):
            if hasattr(self.config, "danbooru_username") and hasattr(self.config, "danbooru_api_key"):
                self.accessible_engines["danbooru"] = Danbooru(self.config.danbooru_username, self.config.danbooru_api_key, self.config.min_score)
                logger.info("Created Danbooru engine")

            # IQDB stuff -> we do the check _first_ since someone might not specify this at all, in which case we do still instantiate it.
            if hasattr(self.config, "skip_iqdb") and self.config.skip_iqdb:
                skip_iqdb = True
            else:
                skip_iqdb = False
            if not skip_iqdb:
                self.accessible_engines["iqdb"] = Iqdb(self.config.min_score)
                logger.info("Created IQDB engine")

        if hasattr(self.config, "saucenao_api_key"):
            self.accessible_engines["saucenao"] = SauceNao(self.config.saucenao_api_key)
            logger.info("Created SauceNao engine")
            if hasattr(self.config, "e621_username") and hasattr(self.config, "app_name") and hasattr(self.config, "version"):
                self.accessible_engines["saucenao"].enable_e621(self.config.e621_username, self.config.app_name, self.config.version)
                logger.info("Activated E621 capabilites on saucenao.")

        self.accessible_engines["paheal"] = Paheal()

        if test_mode:
            self.accessible_engines["dummy"] = Dummy(False)
            self.accessible_engines["fail_dummy"] = Dummy(True)
            self._all_engines.append("dummy")
            self._all_engines.append("fail_dummy")

    async def reverse_search(self, url, callback=None, download=False):
        """
        Reverse searches all accessible boorus for ``url``.

        .. note::
            Callback is a callback function that can be passed in. This can be used to keep track of
            progress for certain methods and functions.

            .. code-block:: python
               :linenos:

                async def callback(engine, rresult):
                    print("This booru was searched: %s", booru)
                    print("These tags were found: %s", rresult.tags)
                    print("This source was found: %s", rresult.source)
                    print("This rating was found: %s", rresult.rating)

                # Callback will be called each time a search finishes.
                rs.reverse_search(url, callback)

        After a reverse search, this method will be called with the engine name that was just searched.

        :param url: The URL to search.
        :type url: str
        :param callback: Callback function.
        :type callback: Optional[function]
        :param download: Run searches on boorus that require a file download. Defaults to False.
        :type download: Optional[bool]
        :return: A :class:`ReverseResult` instance containing your data.
        :rtype: ReverseResult
        """
        tags = set()
        source = set()
        rating = set()
        tasks = []
        logger.info("Creating reverse search engine tasks.")
        for engine in self.accessible_engines:
            if self.accessible_engines[engine].download_required:
                if not download:
                    logger.info("[%s] Downloading files has been disabled. Skipping [%s]", url, engine)
                    continue
            tasks.append(self._gather_reverse_task(engine, url, callback))
        results = await asyncio.gather(*tasks)
        for result in results:
            if isinstance(result, Exception): # pragma: no cover
                logger.warning("[%s] An engine has failed!", url)
                logger.warning("[%s] This may or may not be an issue. Report it on the issue tracker: https://github.com/noirscape/retaggr/issues if the issue persists.", url)
                traceback.print_exc()
                continue
            if result.tags:
                tags.update(result.tags)
            if result.source:
                source.update(result.source)
            if result.rating:
                rating.add(result.rating)
        return ReverseResult(tags, source, rating)

    async def _gather_reverse_task(self, engine, url, callback) -> ImageResult:
        """Underlying method used to run reverse_search more asynchronously."""
        logger.info("[%s] Starting search in [%s] engine", url, engine)
        try:
            result = await self.search_image(engine, url)
        except: # pragma: no cover
            # reverse_search just can't except, it's meant to keep trucking no matter what.
            return ImageResult([], None, None)
        else:
            logger.info("[%s][%s] Found tags: %s", url, engine, result.tags)
            logger.info("[%s][%s] Found source: %s", url, engine, result.source)
            logger.info("[%s][%s] Found rating: %s", url, engine, result.rating)
            if callback:
                logger.info("[%s][%s] Executing callback", url, engine)
                await callback(engine, result)
        logger.info("[%s] Finished searching [%s]", url, engine)
        return result

    async def search_image(self, booru, url):
        r"""Reverse search a booru for ``url``.

        :param booru: Booru to search, this must match a filename in the boorus folder.
        :type booru: str
        :param url: The URL to search.
        :type url: str
        :raises MissingAPIKeysException: Required keys in config object missing.
        :raises NotAValidEngineException: The passed in booru is not a valid booru.
        :return: A :class:`ImageResult` instance containing your data.
        :rtype: ImageResult
        """
        if booru not in self._all_engines:
            raise NotAValidEngineException("%s is not a valid engine", booru)
        if booru not in self.accessible_engines:
            raise MissingAPIKeysException("%s is misisng one or more needed API keys. Check the documentation.")
        return await self.accessible_engines[booru].search_image(url)



class ReverseSearchConfig:
    """
    Configuration object for :class:`ReverseSearch`

    All parameters are prefixed with the specific type needed for reverse searching.

    Check the relevant class for the engine for details on the engine.

    :param danbooru_username: Username on :class:`Danbooru`
    :type danbooru_username: str
    :param danbooru_api_key: API key on :class:`Danbooru`
    :type danbooru_api_key: 
    :param e621_username: Your :class:`E621` username
    :type e621_username: str

    :param app_name: The name of your application (required for the e621 handler in :class:`SauceNao`).
    :type app_name: str
    :param version: The version of your application (required for the e621 handler in :class:`SauceNao`).
    :type version: float

    :param saucenao_api_key: An API key for :class:`SauceNao`.
    :type saucenao_api_key: str

    :param min_score: Minimum search match percentage needed (required for ALL boorus except :class:`Paheal` and :class:`SauceNao`).
    :type min_score: float

    :param skip_iqdb: Don't instantiate the :class:`IQDB` class.
    :type skip_iqdb: bool
    """
    def __init__(self, **kwargs):
        self.__dict__.update(kwargs)


class MissingAPIKeysException(Exception):
    """Raised if a required API key to search an engine is missing."""
    pass

class NotAValidEngineException(Exception):
    """Raised if the passed in engine does not exist."""

class NotAvailableSearchException(Exception):
    """This engine is not capable of searching this option."""

class EngineCooldownException(Exception):
    """This engine is on cooldown."""

class EngineIsDown(Exception):
    """The engine is currently not available (eg. Database issues)."""
